
from typing import TypeVar
from logging import getLogger
import warnings

import numpy as np
from extra_data.read_machinery import select_train_ids


log = getLogger('exdf.data_reduction.ReductionMethod')

train_sel = TypeVar('train_sel')
entry_sel = TypeVar('entry_sel')
index_exp = TypeVar('index_exp')


def is_train_selection(x):
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            select_train_ids([], x)
    except TypeError:
        return False
    else:
        return True


def is_entry_selection(x):
    if isinstance(x, slice):
        return True

    if isinstance(x, list) and all([isinstance(y, (int, bool)) for y in x]):
        return True

    if (
        isinstance(x, np.ndarray) and x.ndim == 1 and
        (np.issubdtype(x.dtype, np.integer) or np.issubdtype(x.dtype, bool))
    ):
        return True

    return False


def is_index_expression(x):
    if isinstance(x, (slice, list)):
        return True

    if isinstance(x, tuple) and all([isinstance(y, (slice, int)) for y in x]):
        return True

    return False


class ReductionMethod(list):
    """A method generating data reduction operations."""

    def __init__(self, data, ap):
        # Methods should always implement their own __init__
        raise NotImplementedError('__init__')

    def _emit(self, *args):
        self.append(args)

        cls = self.__class__
        log.debug(f'<{cls.__module__}.{cls.__qualname__}> emitted {args}')

    def select_trains(
        self,
        source_glob: str,
        trains: train_sel
    ):
        assert isinstance(source_glob, str)
        assert is_train_selection(trains)
        self._emit('select-trains', source_glob, trains)

    def select_entries(
        self,
        source_glob: str,
        index_group: str,
        trains: train_sel,
        entries: entry_sel
    ):
        assert isinstance(source_glob, str)
        assert isinstance(index_group, str)
        assert is_train_selection(trains)
        assert is_entry_selection(entries)
        self._emit('select-entries', source_glob, index_group, trains, entries)

    def select_xtdf(
        self,
        source_glob: str,
        trains: train_sel,
        entries: entry_sel
    ):
        """Slice XTDF data by entry.

        Roughly equivalent to select_entries(source_glob, 'image',
        trains, entries), but only acts on XTDF sources and modifies
        header data structures according to slicing.

        Requires sources to end with :xtdf and have all XTDF keys.

        Args:
            source_glob (str): Source glob pattern.
            trains (train_sel): Train selection.
            entries (entry_sel): Entry selection.
        """
        assert isinstance(source_glob, str)
        assert is_train_selection(trains)
        assert is_entry_selection(entries)
        self._emit('select-xtdf', source_glob, trains, entries)

    def remove_sources(
        self,
        source_glob: str
    ):
        assert isinstance(source_glob, str)
        self._emit('remove-sources', source_glob)

    def remove_keys(
        self,
        source_glob: str,
        key_glob: str
    ):
        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        self._emit('remove-keys', source_glob, key_glob)

    def remove_trains(
        self,
        train_sel: slice
    ):
        """Remove trains from all sources.

        Args:
            trains (train_sel): Train selection.
        """
        assert is_train_selection(train_sel)
        self._emit('remove-trains', '*', train_sel)

    def reshape_keys(
        self,
        source_glob: str,
        key_glob: str,
        shape: tuple
    ):
        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        assert isinstance(shape, tuple)
        assert all([x is None or isinstance(x, (int, slice)) for x in shape])
        self._emit('reshape-keys', source_glob, key_glob, shape)

    def rechunk_keys(
        self,
        source_glob: str,
        key_glob: str,
        chunks: tuple
    ):
        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        assert isinstance(chunks, tuple)
        assert all([x is None or isinstance(x, int) for x in chunks])
        self._emit('rechunk-keys', source_glob, key_glob, chunks)

    def subslice_keys(
        self,
        source_glob: str,
        key_glob: str,
        in_region: index_exp,
        out_region: index_exp = None
    ):
        """Copy only a slice of each key entry.

        By default, the entire input dataset is copied to the output
        dataset and an identical shape is assumed. THis operations
        allows to copy only part of the input data or move it into a
        different place in the output data.

        Args:
            source_glob (str): Source glob pattern.
            key_glob (Str): Key glob pattern,
            in_region: (index_exp): Region in input data to copy from.
            out_region: (index_exp): Region in output data to copy to.
        """

        if out_region is None:
            # TODO: Emit DeprecationWarning
            out_region = in_region

        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        assert is_index_expression(in_region)
        assert is_index_expression(out_region)
        self._emit('subslice-keys', source_glob, key_glob,
                   in_region, out_region)

    def compress_keys(
        self,
        source_glob: str,
        key_glob: str,
        level: int = 1
    ):
        assert isinstance(source_glob, str)
        assert isinstance(key_glob, str)
        assert isinstance(level, int)
        self._emit('compress-keys', source_glob, key_glob, level)

    @staticmethod
    def arguments(ap):
        pass  # Reduction method may not require any arguments.
