
from collections.abc import Iterable, Mapping
from io import StringIO
from packaging.version import Version
import csv

from .method import ReductionMethod


csv.register_dialect(
    'xfel.eu-data-reduction-recipe',
    delimiter='\t',
    lineterminator='\n',
    quoting=csv.QUOTE_NONE,
    skipinitialspace=False,
    strict=True)


class LoadedRecipeMethod(ReductionMethod):
    """Reduction method loaded from a recipe.

    Reduction methods loaded from a recipe use this type to allow direct
    instantation and carry class identity of loaded methods.
    """

    def __init__(self, identity):
        super(ReductionMethod, self).__init__()
        self.identity = identity


def dump(methods_or_ops, fp, headers=[]):
    """Serialize reduction operations to a recipe.

    Args:
        methods (Mapping or Iterable of ReductionMethod): Reduction
            methods to serialize structured into section when passed as
            a mapping or a simple list when passed as just an iterable.
        generated_by (str, optional): Software that generated this file.
        headers (Iterable of str, optional): Additional header lines.

    Returns:
        None
    """

    writer = csv.writer(fp, dialect='xfel.eu-data-reduction-recipe')
    writer.writerow(['xfel.eu data-reduction-recipe v1.0'])

    for header in headers:
        writer.writerow([f'# {header}'])

    if isinstance(methods_or_ops, Mapping):
        for name, method in methods_or_ops.items():
            cls = method.__class__
            fp.write(writer.dialect.lineterminator)

            if isinstance(method, LoadedRecipeMethod):
                writer.writerow([f'- {method.identity}'])
            else:
                writer.writerow(
                    [f'- {name} <{cls.__module__}.{cls.__qualname__}>'])

            for op in method:
                writer.writerow(op)

    elif isinstance(methods_or_ops, Iterable):
        fp.write(writer.dialect.lineterminator)
        writer.writerows(methods_or_ops)

    else:
        raise ValueError('ops must either be mapping or iterable')


def dumps(methods, **kwargs):
    f = StringIO()
    dump(methods, f, **kwargs)
    return f.getvalue()


def load(fp):
    reader = csv.reader(fp, dialect='xfel.eu-data-reduction-recipe')

    line = next(reader)[0].strip()
    if line[:line.rfind(' ')] != 'xfel.eu data-reduction-recipe':
        raise ValueError('missing file definition, should be '
                         '`xfel.eu data-reduction-recipe <version>`')

    Version(line[line.rfind(' ')+1:])

    result = LoadedRecipeMethod('n/a')
    headers = []

    in_header = True
    cur_method = None

    for row in reader:
        if not row:
            continue

        elif row[0].startswith('#'):
            if in_header:
                headers.append(row[0].lstrip('# '))

        elif row[0].startswith('-'):
            in_header = False

            method_header = row[0].lstrip('- ')

            try:
                cur_method = method_header[:method_header.index(' ')]
            except ValueError:
                cur_method = method_header

            if isinstance(result, LoadedRecipeMethod):
                # Change result to dict.
                if result:
                    result = {'n/a': result}
                else:
                    result = {}

            result[cur_method] = LoadedRecipeMethod(method_header)

        else:
            in_header = False

            if cur_method is not None:
                cur_ops = result[cur_method]
            else:
                cur_ops = result

            cur_ops.append(tuple(row))

    return result, headers


def loads(recipe):
    return load(StringIO(recipe))
